#!/usr/bin/python3
# -*- coding: utf-8 -*-

# Slimbook Service
# Copyright (C) 2022 Slimbook 
# In case you modify or redistribute this code you must keep the copyright line above.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import slimbook.info
import slimbook.smbios

import os, codecs, json
import subprocess
import locale
import gettext
import requests
import re
import signal
import sys

LANGDOMAIN = "slimbook"
LANGDIR = "/usr/share/slimbook/locale-langpack"

try:
    current_locale, encoding = locale.getdefaultlocale()
    language = gettext.translation(LANGDOMAIN, LANGDIR, [current_locale])
    language.install()
    _ = language.gettext
except Exception as e:
    print(e, file = sys.stderr)
    _ = str

# will be handy to have this variable already here
xdg_current_desktop = os.environ.get("XDG_CURRENT_DESKTOP")

SLB_EVENT_QC71_SILENT_MODE_CHANGED = 0x00
SLB_EVENT_QC71_SILENT_MODE_ON = 0x01
SLB_EVENT_QC71_SILENT_MODE_OFF = 0x02

SLB_EVENT_QC71_SUPER_LOCK_CHANGED = 0x03
SLB_EVENT_QC71_SUPER_LOCK_ON = 0x04
SLB_EVENT_QC71_SUPER_LOCK_OFF = 0x05

SLB_EVENT_QC71_SILENT_MODE = 0x06
SLB_EVENT_QC71_NORMAL_MODE = 0x07
SLB_EVENT_QC71_PERFORMANCE_MODE = 0x08
SLB_EVENT_QC71_DYNAMIC_MODE = 0x09

#this events are shared on several platforms and no longer are qc71 exclusive
SLB_EVENT_TOUCHPAD_CHANGED = 0x0100
SLB_EVENT_TOUCHPAD_ON = 0x0200
SLB_EVENT_TOUCHPAD_OFF = 0x0300
SLB_EVENT_WEBCAM_CHANGED = 0x0400
SLB_EVENT_WEBCAM_ON = 0x0500
SLB_EVENT_WEBCAM_OFF = 0x0600

SLB_EVENT_ENERGY_SAVER_MODE = 0x0700
SLB_EVENT_BALANCED_MODE = 0x0800
SLB_EVENT_PERFORMANCE_MODE = 0x0900

SLB_EVENT_AC_OFFLINE = 0x1000
SLB_EVENT_AC_ONLINE = 0x1001

SLB_EVENT_QC71_INPUT_LOADED = 0x2000
SLB_EVENT_QC71_INPUT_UNLOADED = 0x2001

SLB_EVENT_UPOWER_POWER_EVENT = 0x4000
SLB_EVENT_UPOWER_POWER_SAVER = 0x4001
SLB_EVENT_UPOWER_BALANCED    = 0x4002
SLB_EVENT_UPOWER_PERFORMANCE = 0x4003

if xdg_current_desktop == "KDE":
    SLB_EVENT_DATA = {
        SLB_EVENT_QC71_SILENT_MODE_ON : [_("Silent Mode enabled"),"battery-profile-powersave-symbolic"],
        SLB_EVENT_QC71_SILENT_MODE_OFF : [_("Silent Mode disabled"),"battery-profile-balanced-symbolic"],
        SLB_EVENT_QC71_SILENT_MODE_CHANGED : [_("Silent Mode changed"),"battery-profile-balanced-symbolic"],

        SLB_EVENT_QC71_SUPER_LOCK_ON : [_("Super Key Lock enabled"),"keyboard-layout"],
        SLB_EVENT_QC71_SUPER_LOCK_OFF : [_("Super Key Lock disabled"),"keyboard-layout"],
        SLB_EVENT_QC71_SUPER_LOCK_CHANGED : [_("Super Key Lock changed"),"keyboard-layout"],

        SLB_EVENT_QC71_SILENT_MODE : [_("Silent Mode"),"battery-profile-powersave-symbolic"],
        SLB_EVENT_QC71_NORMAL_MODE : [_("Normal Mode"),"battery-profile-balanced-symbolic"],
        SLB_EVENT_QC71_PERFORMANCE_MODE : [_("Performance Mode"),"battery-profile-performance-symbolic"],
        SLB_EVENT_QC71_DYNAMIC_MODE : [_("Dynamic Mode"),"battery-profile-powersave-symbolic"],

        SLB_EVENT_TOUCHPAD_ON : [_("Touchpad enabled"),"input-touchpad-on-symbolic"],
        SLB_EVENT_TOUCHPAD_OFF : [_("Touchpad disabled"),"input-touchpad-off-symbolic"],
        SLB_EVENT_TOUCHPAD_CHANGED : [_("Touchpad changed"),"touchpad_enabled"],
        SLB_EVENT_WEBCAM_CHANGED : [_("Webcam changed"),"camera-on"],
        SLB_EVENT_WEBCAM_ON : [_("Webcam enabled"),"camera-on"],
        SLB_EVENT_WEBCAM_OFF : [_("Webcam disabled"),"camera-off"],

        SLB_EVENT_ENERGY_SAVER_MODE : [_("Energy Saver"),"battery-profile-powersave-symbolic"],
        SLB_EVENT_BALANCED_MODE : [_("Balanced"),"battery-profile-balanced-symbolic"],
        SLB_EVENT_PERFORMANCE_MODE : [_("Performance"),"battery-profile-performance-symbolic"]
    }

else:
    SLB_EVENT_DATA = {
        SLB_EVENT_QC71_SILENT_MODE_ON : [_("Silent Mode enabled"),"power-profile-power-saver-symbolic"],
        SLB_EVENT_QC71_SILENT_MODE_OFF : [_("Silent Mode disabled"),"power-profile-balanced-symbolic"],
        SLB_EVENT_QC71_SILENT_MODE_CHANGED : [_("Silent Mode changed"),"power-profile-balanced-symbolic"],

        SLB_EVENT_QC71_SUPER_LOCK_ON : [_("Super Key Lock enabled"),"preferences-system-privacy-symbolic"],
        SLB_EVENT_QC71_SUPER_LOCK_OFF : [_("Super Key Lock disabled"),"preferences-system-privacy-symbolic"],
        SLB_EVENT_QC71_SUPER_LOCK_CHANGED : [_("Super Key Lock changed"),"preferences-system-privacy-symbolic"],

        SLB_EVENT_QC71_SILENT_MODE : [_("Silent Mode"),"power-profile-power-saver-symbolic"],
        SLB_EVENT_QC71_NORMAL_MODE : [_("Normal Mode"),"power-profile-balanced-symbolic"],
        SLB_EVENT_QC71_PERFORMANCE_MODE : [_("Performance Mode"),"power-profile-performance-symbolic"],
        SLB_EVENT_QC71_DYNAMIC_MODE : [_("Dynamic Mode"),"power-profile-power-saver-symbolic"],

        SLB_EVENT_TOUCHPAD_ON : [_("Touchpad enabled"),"input-touchpad-symbolic"],
        SLB_EVENT_TOUCHPAD_OFF : [_("Touchpad disabled"),"input-touchpad-symbolic"],
        SLB_EVENT_TOUCHPAD_CHANGED : [_("Touchpad changed"),"input-touchpad-symbolic"],
        SLB_EVENT_WEBCAM_CHANGED : [_("Webcam changed"),"preferences-system-privacy-symbolic"],
        SLB_EVENT_WEBCAM_ON : [_("Webcam enabled"),"preferences-system-privacy-symbolic"],
        SLB_EVENT_WEBCAM_OFF : [_("Webcam disabled"),"preferences-system-privacy-symbolic"],

        SLB_EVENT_ENERGY_SAVER_MODE : [_("Energy Saver"),"power-profile-power-saver-symbolic"],
        SLB_EVENT_BALANCED_MODE : [_("Balanced"),"power-profile-balanced-symbolic"],
        SLB_EVENT_PERFORMANCE_MODE : [_("Performance"),"power-profile-performance-symbolic"]
    }

PARAMS = {
            'first-time': True,
            'version': '',
            'autostart': True,
            'theme': 'light',
            'show': True,
            'notifications' : True,
            'trackpad-lock' : True,
            'power-profile' : True,
            'ac-notifications' : True
            }

POWER_PROFILE_POWER_SAVER = "power-saver"
POWER_PROFILE_BALANCED = "balanced"
POWER_PROFILE_PERFORMANCE = "performance"

POWER_PROFILE_NAME = {
    slimbook.info.SLB_QC71_PROFILE_UNKNOWN : "unknown",
    slimbook.info.SLB_QC71_PROFILE_ENERGY_SAVER : POWER_PROFILE_POWER_SAVER,
    slimbook.info.SLB_QC71_PROFILE_BALANCED : POWER_PROFILE_BALANCED,
    slimbook.info.SLB_QC71_PROFILE_PERFORMANCE : POWER_PROFILE_PERFORMANCE
}

TUNED_PROFILE = {
    POWER_PROFILE_POWER_SAVER : "powersave",
    POWER_PROFILE_BALANCED : "balanced-battery",
    POWER_PROFILE_PERFORMANCE : "throughput-performance"
    }
    
POWER_NAME_TO_EVENT = {
    POWER_PROFILE_POWER_SAVER : SLB_EVENT_UPOWER_POWER_SAVER,
    POWER_PROFILE_BALANCED : SLB_EVENT_UPOWER_BALANCED,
    POWER_PROFILE_PERFORMANCE : SLB_EVENT_UPOWER_PERFORMANCE,
    
    #tuned names
    TUNED_PROFILE[POWER_PROFILE_POWER_SAVER] : SLB_EVENT_UPOWER_POWER_SAVER,
    TUNED_PROFILE[POWER_PROFILE_BALANCED] : SLB_EVENT_UPOWER_BALANCED,
    TUNED_PROFILE[POWER_PROFILE_PERFORMANCE] : SLB_EVENT_UPOWER_PERFORMANCE
}

OPT_TRACKPAD_LOCK    = "trackpad-lock"
OPT_POWER_PROFILE    = "power-profile"
OPT_AC_NOTIFICATIONS = "ac-notifications"

CMD_LOAD_SETTINGS = "cmd-load"

QC71_DOUBLE_PROFILE = [slimbook.info.SLB_MODEL_PROX, slimbook.info.SLB_MODEL_EXECUTIVE]
QC71_TRIPLE_PROFILE = [slimbook.info.SLB_MODEL_TITAN, slimbook.info.SLB_MODEL_HERO, slimbook.info.SLB_MODEL_EVO, slimbook.info.SLB_MODEL_CREATIVE]

QC71_DOUBLE_PROFILE_FROM_UPOWER = {
    SLB_EVENT_UPOWER_POWER_SAVER : slimbook.info.SLB_QC71_PROFILE_ENERGY_SAVER,
    SLB_EVENT_UPOWER_BALANCED : slimbook.info.SLB_QC71_PROFILE_BALANCED,
    SLB_EVENT_UPOWER_PERFORMANCE : slimbook.info.SLB_QC71_PROFILE_BALANCED
}

QC71_TRIPLE_PROFILE_FROM_UPOWER = {
    SLB_EVENT_UPOWER_POWER_SAVER : slimbook.info.SLB_QC71_PROFILE_ENERGY_SAVER,
    SLB_EVENT_UPOWER_BALANCED : slimbook.info.SLB_QC71_PROFILE_BALANCED,
    SLB_EVENT_UPOWER_PERFORMANCE : slimbook.info.SLB_QC71_PROFILE_PERFORMANCE
}

QC71_DOUBLE_PROFILE_TO_NOTIFICATION = {
    slimbook.info.SLB_QC71_PROFILE_ENERGY_SAVER : SLB_EVENT_QC71_SILENT_MODE_ON,
    slimbook.info.SLB_QC71_PROFILE_BALANCED : SLB_EVENT_QC71_SILENT_MODE_OFF
}

QC71_TRIPLE_PROFILE_TO_NOTIFICATION = {
    slimbook.info.SLB_QC71_PROFILE_ENERGY_SAVER : SLB_EVENT_ENERGY_SAVER_MODE,
    slimbook.info.SLB_QC71_PROFILE_BALANCED : SLB_EVENT_BALANCED_MODE,
    slimbook.info.SLB_QC71_PROFILE_PERFORMANCE : SLB_EVENT_PERFORMANCE_MODE
}

#set a default dark theme for kde
if xdg_current_desktop == "KDE":
    PARAMS['theme'] = 'dark'

APP = 'slimbook'
VERSION = '1.0'
APPCONF = APP + '.conf'
APPDATA = APP + '.data'
APPNAME = _('Slimbook Service')
CONFIG_DIR = os.path.join(os.path.expanduser('~'), '.config')
CONFIG_APP_DIR = os.path.join(CONFIG_DIR, APP)
CONFIG_FILE = os.path.join(CONFIG_APP_DIR, APPCONF)
DATA_FILE = os.path.join(CONFIG_APP_DIR, APPDATA)
AUTOSTART_DIR = os.path.join(CONFIG_DIR, 'autostart')
FILE_AUTO_START = os.path.join(AUTOSTART_DIR,
                               'slimbook-client-autostart.desktop')
                               
SLB_FEED_URL = "https://github.com/Slimbook-Team/slimbook-notifications-feed/raw/main/slb-rss-{0}.xml"
SLB_CACHE_PATH = os.path.expanduser("~/.cache/slimbook-service/")

SLB_IPC_PATH     = "/var/run/slimbook-service.socket"
SLB_IPC_CTL_PATH = "/var/run/slimbook-service-ctl.socket"

def is_package():
    return os.path.abspath(os.path.dirname(__file__)).startswith('/usr')

if is_package():
    APPDIR = '/usr/share/slimbook'
else:
    APPDIR = os.path.abspath(os.path.join('..', os.path.dirname(os.path.realpath(__file__))))

LANGDIR = os.path.join(APPDIR, 'locale-langpack')
ICONDIR = os.path.join(APPDIR, 'icons')
FILE_AUTO_START_ORIG = os.path.join(APPDIR,
                                    'slimbook-client-autostart.desktop')

ICON = 'slimbook-be1ofus-light'

STATUS_ICON = {}
STATUS_ICON['light'] = 'slimbook-status-active-light'
STATUS_ICON['light-attention'] = 'slimbook-status-attention-light'
STATUS_ICON['dark'] = 'slimbook-status-active-dark'
STATUS_ICON['dark-attention'] = 'slimbook-status-attention-dark'

INFO_UPTIME = _("Uptime")
INFO_MEM = _("Memory Free/Total")
INFO_MEM_DEVICE = _("Memory device")
INFO_UMA = _("UMA")
INFO_DISK_DEVICE = _("Disk Free/Total")
INFO_KERNEL = _("Kernel")
INFO_OS = _("OS")
INFO_DESKTOP = _("Desktop")
INFO_SESSION = _("Session")
INFO_PRODUCT = _("Product")
INFO_SERIAL = _("Serial")
INFO_BIOS = _("Bios Version")
INFO_EC = _("EC Version")
INFO_BOOT = _("Boot Mode")
INFO_SB = _("Secure Boot")
INFO_CPU = _("CPU")
INFO_TDP = _("TDP")
INFO_GPU = _("GPU")
INFO_MODULE = _("Module loaded")
INFO_FN_LOCK = _("Fn Lock")
INFO_SUPER_LOCK = _("Super Lock")
INFO_SILENT_MODE = _("Silent Mode")
INFO_TURBO_MODE = _("Turbo Mode")
INFO_PROFILE = _("Profile")

INFO_YES = _("Yes")
INFO_NO = _("No")

CONTROL_PANEL_PATH = "/usr/bin/slimbook-controlpanel"

class Configuration(object):
    def __init__(self):
        self.params = PARAMS
        self.read()

    def get(self, key):
        try:
            return self.params[key]
        except KeyError as e:
            print(e)
            self.params[key] = PARAMS[key]
            return self.params[key]

    def set(self, key, value):
        self.params[key] = value

    def reset(self):
        if os.path.exists(CONFIG_FILE):
            os.remove(CONFIG_FILE)
        self.params = PARAMS
        self.save()

    def set_defaults(self):
        self.params = PARAMS
        self.save()

    def read(self):
        try:
            f = codecs.open(CONFIG_FILE, 'r', 'utf-8')
        except IOError as e:
            print(e)
            self.save()
            f = codecs.open(CONFIG_FILE, 'r', 'utf-8')
        try:
            self.params = json.loads(f.read())
        except ValueError as e:
            print(e)
            self.save()
        f.close()

    def save(self):
        if not os.path.exists(CONFIG_APP_DIR):
            os.makedirs(CONFIG_APP_DIR)

        f = codecs.open(CONFIG_FILE, 'w', 'utf-8')
        f.write(json.dumps(self.params, separators=(",\n", ": ")))
        f.close()

def _read_file(file):
    f = open(file,"r")
    data = f.readlines()
    f.close()
    
    return data

def _get_pciid(vendor,device):
    ret=[vendor,device]
    
    f=open("/usr/share/misc/pci.ids","r")
    lines=f.readlines()
    f.close()
    
    vm = False
    
    for line in lines:
        if not line[0]=="\t":
            if vm:
                break
            
            if vendor == line[:4]:
                ret[0]=line[6:].strip()
                vm = True
        else:
            if vm:
                if device == line[1:5]:
                    ret[1]=line[6:].strip()
    
    return ret
    
def _get_cpu():
    ret = []

    f = open("/proc/cpuinfo","r")
    lines = f.readlines()
    f.close()
    
    model = ""
    cpus = {}
    
    for line in lines:
        tmp = line.split(":")
        if (len(tmp) < 2):
            continue
        
        key = tmp[0].strip()
        value = tmp[1].strip()
        
        if (key == "model name"):
            model = value
    
        if (key == "physical id"):
            if not value in cpus:
                cpus[value] = (model,1)
            else:
                c,count = cpus[value]
                cpus[value] = (c,count + 1)
    
    for k in cpus:
        ret.append(cpus[k][0] + " x " + str(cpus[k][1]))
     
    return ret

def _get_gpu():
    gpus = []
    
    for n in range(8):
        gpu_path = "/sys/class/drm/card{0}".format(n)
        if os.path.exists(gpu_path):
            vendor = _read_file(gpu_path+"/device/vendor")[0].strip()
            device = _read_file(gpu_path+"/device/device")[0].strip()
            vendor = vendor[2:]
            device = device[2:]
            
            vendor,device = _get_pciid(vendor,device)
            
            gpus.append("{0} {1}".format(vendor,device))
    
    return gpus
    
def get_system_info():
    info = []
    
    sb_platform = slimbook.info.get_platform()
    
    uptime = slimbook.info.uptime()
    h = int(uptime / 3600)
    m = int((uptime / 60) % 60)
    s = uptime % 60
    
    txt = "{0}h {1}m {2}s".format(h,m,s)
    info.append([INFO_UPTIME,txt])
    
    try:
        data = _read_file("/proc/version")
        info.append([INFO_KERNEL,data[0].strip().split()[2]])
    except:
        pass
    
    serial=""
    memory_devices = []
    disk_devices = []
    memory = ""
    uma = ""
    tdp = ""
    
    is_module = INFO_NO
    fn_lock = ""
    super_lock = ""
    silent_mode = ""
    profile = ""
    
    tmp = []
    
    try:
        tmp = subprocess.getstatusoutput("slimbookctl info")[1]
        tmp = tmp.split('\n')
    except:
        pass
    
    for line in tmp:
        pair = line.split(':')
        if (len(pair) == 2):
            key = pair[0]
            value = pair[1]
            
            if (key == "serial"):
                serial = value
            
            if (key == "memory device"):
                memory_devices.append(value)
            
            if (key == "disk free/total"):
                idx = value.find(" ")
                
                disk_devices.append(value[:idx] + "    " + value[idx:])
            
            if (key == "memory free/total"):
                memory = value
            
            if (key == "module loaded"):
                is_module = value.capitalize()
             
            if (key == "fn lock"):
                fn_lock = value.capitalize()
             
            if (key == "super key lock"):
                super_lock = value.capitalize()
             
            if (key == "silent mode"):
                silent_mode = value.capitalize()

            if (key == "profile"):
                profile = value.capitalize()
            
            # warning, this case may change in the future
            if (key == "UMA Framebuffer"):
                uma = value
   
            if (key == "TDP"):
                tdp = value
            
            if (key.startswith("TDP sustained")):
                tdp = value
            
    info.append([INFO_MEM,memory])
    
    for d in disk_devices:
        info.append([INFO_DISK_DEVICE,d])
    
    try:
        if (os.path.exists("/sys/firmware/efi")):
            info.append([INFO_BOOT,"UEFI"])
            sb = False
            SB_VAR = "/sys/firmware/efi/efivars/SecureBoot-8be4df61-93ca-11d2-aa0d-00e098032b8c"
            if (os.path.exists(SB_VAR)):
                f = open(SB_VAR,"rb")
                var = list(f.read())
                if (var[4] == 1):
                    sb = True
                f.close()

            if sb:
                info.append([INFO_SB,INFO_YES])
            else:
                info.append([INFO_SB,INFO_NO])
        else:
            info.append([INFO_BOOT,"Legacy"])
    except:
        pass
    
    try:
        if (os.path.exists("/usr/lib/os-release")):
            f = open("/usr/lib/os-release","rt")
            lines = f.readlines()
            f.close()

            name = None
            version = None

            for line in lines:
                tmp = line.strip().split('=')

                if (len(tmp) > 1):
                    if (tmp[0] == "NAME"):
                        name = tmp[1].strip("\"")
                    if (tmp[0] == "VERSION"):
                        version = tmp[1].strip("\"")
            if (name and version):
                info.append([INFO_OS,name + " " + version])
    except:
        pass

    try:
        info.append([INFO_DESKTOP, os.environ["XDG_CURRENT_DESKTOP"].replace(":",", ")])
    except:
        pass

    try:
        info.append([INFO_SESSION, os.environ["XDG_SESSION_TYPE"]])
    except:
        pass
    
    try:
        data = _read_file("/sys/class/dmi/id/product_name")
        info.append([INFO_PRODUCT,data[0].strip()])
    except:
        pass
    
    try:
        data = _read_file("/sys/class/dmi/id/bios_version")
        info.append([INFO_BIOS,data[0].strip()])
    except:
        pass
    
    try:
        data = _read_file("/sys/class/dmi/id/ec_firmware_release")
        info.append([INFO_EC,data[0].strip()])
    except:
        pass
    
    info.append([INFO_SERIAL,serial])

    try:
        for cpu in _get_cpu():
            info.append([INFO_CPU, cpu])
    except:
        pass
    
    info.append([INFO_TDP,tdp])
    
    try:
        for gpu in _get_gpu():
            info.append([INFO_GPU,gpu])
    except:
        pass
        
    for m in memory_devices:
        info.append([INFO_MEM_DEVICE,m])
        
    info.append([INFO_UMA,uma])
        
    if (sb_platform != 0 ):
        info.append([INFO_MODULE,is_module])
        
        if (sb_platform == slimbook.info.SLB_PLATFORM_QC71):
            info.append([INFO_FN_LOCK,fn_lock])
            info.append([INFO_SUPER_LOCK,super_lock])
            info.append([INFO_PROFILE,profile])
    
    return info

def get_lang():
    lang = locale.getlocale()[0]
    lang = lang.split("_")[0]
    
    if not lang in ["en","es"]:
        lang = "en"
        
    return lang

def download_feed():
    os.makedirs(SLB_CACHE_PATH,exist_ok = True)

    lang = get_lang()
    #print(SLB_FEED_URL.format(lang))
    r = requests.get(SLB_FEED_URL.format(lang), allow_redirects = True)

    path = SLB_CACHE_PATH + "/sb-rss.xml"
    f = open(path,"wb")
    f.write(r.content)
    f.close()

def report_proc(self, glib_cb, cb, report_type):
        proc = subprocess.Popen(["slimbookctl", report_type], stdout= subprocess.PIPE, stderr= subprocess.PIPE)
        cb_args = [False, "", ""]

        while(proc.poll() == None):
            glib_cb(cb, cb_args)  

        proc.wait()

        ret = proc.poll()

        if(ret != 0):            
            match ret:
                #should never happen
                case 1:
                    cb_args[1] = "error"
                case -1:
                    cb_args[1] = "terminal disconnected (SIGHUP)"
                #should never happen
                case -2:
                    cb_args[1] = "process stopping by user(SIGINT)"
                #should never happen unless OOM?
                case -9:
                    cb_args[1] = "process killed (SIGKILL)"
                case -15:
                    cb_args[1] = "process terminated (SIGTERM)"
                #should never happen
                case -19:
                    cb_args[1] = "process stopped (SIGSTOP)"
                #should never happen
                case -20:
                    cb_args[1] = "process stopped by user (SIGTSTP)"

        try:
            o = proc.communicate(timeout = 5) 
        except TimeoutExpired:
            proc.kill()
            o = proc.communicate()
            
        if re.search(r"\/.*", o[0].decode("utf-8")):
            cb_args.pop(2)
            path = re.search(r"\/.*", o[0].decode("utf-8")).group(0)
            cb_args.append(path)

        cb_args.pop(0)
        cb_args.insert(0, True)

        glib_cb(cb, cb_args)  



#!/usr/bin/python3
# -*- coding: utf-8 -*-

# Slimbook Service
# Copyright (C) 2022 Slimbook
# In case you modify or redistribute this code you must keep the copyright line above.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import common
import touchpad

import slimbook.info
import slimbook.qc71

from gi.repository import GLib, Gio

import zmq
import evdev
import pyudev

import subprocess
from datetime import datetime
import os
import sys
import logging
import threading
import queue
import time

logger = logging.getLogger("slimbook.service")
logging.basicConfig(format='%(levelname)s-%(message)s')
logger.setLevel(logging.INFO)

context = zmq.Context()
socket_out = context.socket(zmq.PUB)
socket_out.bind("ipc://{0}".format(common.SLB_IPC_PATH))
os.chmod(common.SLB_IPC_PATH, 0o777)

socket_ctl = context.socket(zmq.REP)
#socket_ctl.setsockopt_string(zmq.SUBSCRIBE, "")
#socket_ctl.setsockopt(zmq.SUBSCRIBE, b'')
socket_ctl.bind("ipc://{0}".format(common.SLB_IPC_CTL_PATH))
os.chmod(common.SLB_IPC_CTL_PATH, 0o777)

slb_events = queue.Queue()

# whenever platform driver is loaded or not
module_loaded = False

settings = {
    common.OPT_TRACKPAD_LOCK: True,
    common.OPT_POWER_PROFILE: True,
    common.OPT_AC_NOTIFICATIONS: True
}

def set_power_profile(profile):
    #ToDo: refactor this using Dbus instead
    if (settings[common.OPT_POWER_PROFILE]):
        if (os.path.exists("/usr/bin/powerprofilesctl")):
            subprocess.run(["/usr/bin/powerprofilesctl","set",profile])
        elif (os.path.exists("/usr/bin/tuned-adm")):
            subprocess.run(["/usr/bin/tuned-adm","profile",common.TUNED_PROFILE[profile]])

def get_udev_ac_status(device):
    try:
        ps_type = device.get("POWER_SUPPLY_TYPE")
        ps_online = device.get("POWER_SUPPLY_ONLINE")
        
        if (ps_type and ps_online):
            if (ps_type == "Mains"):
                return int(ps_online)
    except:
        pass
    
    return -1

def upower_change(mode):
    if mode:
        event = common.POWER_NAME_TO_EVENT[mode]
        slb_events.put(event)
    
def upower_hndlr(dbus_proxy, properties_changed, properties_removed):
    props = properties_changed.unpack()
    upower_change(props.get("ActiveProfile"))
    
def upower_worker():
    upower_proxy = Gio.DBusProxy.new_for_bus_sync(
            bus_type = Gio.BusType.SYSTEM,
            flags = Gio.DBusProxyFlags.NONE,
            info = None,
            name = 'org.freedesktop.UPower.PowerProfiles',
            object_path = '/org/freedesktop/UPower/PowerProfiles',
            interface_name = 'org.freedesktop.UPower.PowerProfiles',
            cancellable = None)

    upower_proxy.connect('g-properties-changed', upower_hndlr)
    upower_change(upower_proxy.get_cached_property("ActiveProfile").unpack())

    ctx = GLib.MainContext.default()
    
    while (ctx.iteration(True)):
        pass

def udev_worker():
    context = pyudev.Context()
    last_ac_status = -1

    for device in context.list_devices(subsystem="power_supply"):
        status = get_udev_ac_status(device)

        if (status >=0):
            last_ac_status = status
            logger.info("AC status:{0}".format(status))
            slb_events.put(common.SLB_EVENT_AC_OFFLINE + status)
    
    for device in context.list_devices(subsystem="input"):
        if device.get("ID_PATH") == "platform-qc71_laptop":
            if (device.get("DEVNAME")):
                slb_events.put(common.SLB_EVENT_QC71_INPUT_LOADED)

    monitor = pyudev.Monitor.from_netlink(context)
    #monitor.filter_by('power_supply')
    for device in iter(monitor.poll, None):
        if device.subsystem == "power_supply":
            status = get_udev_ac_status(device)
            if (status >=0 and status != last_ac_status):
                last_ac_status = status
                logger.info("AC status:{0}".format(status))
                slb_events.put(common.SLB_EVENT_AC_OFFLINE + status)
        elif device.subsystem == "input":
            if (device.get("ID_PATH") == "platform-qc71_laptop" and device.get("DEVNAME")):
                if (device.action == "add"):
                    slb_events.put(common.SLB_EVENT_QC71_INPUT_LOADED)
                else:
                    slb_events.put(common.SLB_EVENT_QC71_INPUT_UNLOADED)
                
def zmq_worker():
    
    while True: 
        if (socket_ctl.poll(timeout = 100) == 0):
            continue
        data = socket_ctl.recv_json()
        
        cmd = data.get("cmd")
        
        if (cmd and cmd == common.CMD_LOAD_SETTINGS):
            
            keys = data.get("settings")
            if (keys):
                logger.info("Updating settings...")
                for k in keys:
                    logger.info("{0}={1}".format(k,keys[k]))
                    settings[k] = keys[k]
                
        
        socket_ctl.send_json({})
    
def keyboard_worker():
    
    device_path = "/dev/input/by-path/platform-i8042-serio-0-event-kbd"
    # work around for buggy dmi info
    try:
        device_path = slimbook.info.keyboard_device()
    except:
        pass
        
    device = evdev.InputDevice(device_path)
    
    state = {}
    
    for event in device.read_loop():
        if (event.type == evdev.ecodes.EV_MSC):
        
            last = state.get(event.value)
            
            if (last == 1):
                state[event.value] = 0
                continue
            else:
                state[event.value] = 1
            
            if (event.value == slimbook.info.SLB_SCAN_QC71_SUPER_LOCK):
                slb_events.put(common.SLB_EVENT_QC71_SUPER_LOCK_CHANGED)
            
            elif (event.value == slimbook.info.SLB_SCAN_QC71_SILENT_MODE and module_loaded == False):
                logger.debug("qc71 performance change requested (i8042)")
                slb_events.put(common.SLB_EVENT_QC71_SILENT_MODE_CHANGED)
            
            elif (event.value == slimbook.info.SLB_SCAN_TOUCHPAD_SWITCH):
                slb_events.put(common.SLB_EVENT_TOUCHPAD_CHANGED)
    
            elif (event.value == slimbook.info.SLB_SCAN_ENERGY_SAVER_MODE):
                slb_events.put(common.SLB_EVENT_ENERGY_SAVER_MODE)
                
            elif (event.value == slimbook.info.SLB_SCAN_BALANCED_MODE):
                slb_events.put(common.SLB_EVENT_BALANCED_MODE)
                
            elif (event.value == slimbook.info.SLB_SCAN_PERFORMANCE_MODE):
                slb_events.put(common.SLB_EVENT_PERFORMANCE_MODE)

def qc71_module_worker():
    logger.debug("qc71 keyboard worker start")
    device = evdev.InputDevice(slimbook.info.module_device())
    try:
        for event in device.read_loop():
            if (event.type == evdev.ecodes.EV_KEY):
                if (event.value == 1 and event.code == evdev.ecodes.KEY_FN_F2):
                    slb_events.put(common.SLB_EVENT_QC71_SUPER_LOCK_CHANGED)
                elif (event.value == 1 and event.code == evdev.ecodes.KEY_FN_F5):
                    logger.debug("qc71 performance change requested")
                    slb_events.put(common.SLB_EVENT_QC71_SILENT_MODE_CHANGED)
                elif (event.value == 1 and event.code == evdev.ecodes.KEY_FN_F12):
                    slb_events.put(common.SLB_EVENT_WEBCAM_CHANGED)
    except:
        pass
    logger.info("qc71 keyboard thread end")
    
def send_notify(code):
    dt = datetime.now()
    ts = datetime.timestamp(dt)
    data = {"code": code, "timestamp": ts}
    socket_out.send_json(data)
    
def main():
    logger.info("Slimbook service")

    zmq_thread = threading.Thread(
            name='slimbook.service.zmq', target=zmq_worker)
    zmq_thread.start()
    
    udev_thread = threading.Thread(
            name='slimbook.service.udev', target=udev_worker)
    udev_thread.start()
    
    upower_thread = threading.Thread(
            name='slimbook.service.upower', target=upower_worker)
    upower_thread.start()
    
        
    tpad = touchpad.Touchpad()
    if (tpad.valid()):
        tpad_mode_name = {touchpad.Touchpad.MODE_HIDRAW:"hidraw",touchpad.Touchpad.MODE_EVDEV:"evdev"}
        logger.info("Found a touchpad device of type {0}".format(tpad_mode_name[tpad.mode]))
    
    keyboard_platforms = [slimbook.info.SLB_PLATFORM_Z16,slimbook.info.SLB_PLATFORM_HMT16]
    
    model = slimbook.info.get_model()
    platform = slimbook.info.get_platform()
    family = slimbook.info.get_family()
    
    power_profiles = slimbook.info.get_performance_profiles()

    logger.info("platform:{0:04x}".format(platform))
    logger.info("model:{0:04x}".format(model))
    logger.info("power profiles:{0}".format(power_profiles))
    
    if (model == slimbook.info.SLB_MODEL_UNKNOWN):
        product = slimbook.info.product_name().lower()
        vendor = slimbook.info.board_vendor().lower()
        
        if (product.startswith("excalibur")):
            # work-around for buggy dmi data
            model = slimbook.info.SLB_MODEL_EXCALIBUR
            platform = slimbook.info.SLB_PLATFORM_Z16
        else:
            logger.warning("Unknown model:")
            logger.warning("Product:[{0}]".format(slimbook.info.product_name()))
            logger.warning("Vendor:[{0}]".format(slimbook.info.board_vendor()))
    
    module_loaded = slimbook.info.is_module_loaded()
    
    if (platform == slimbook.info.SLB_PLATFORM_QC71):
        qc71_keyboard_thread = threading.Thread(
            name='slimbook.service.qc71.keyboard', target=keyboard_worker)
        qc71_keyboard_thread.start()
            
        if (module_loaded):
            logger.info("Setting qc71 manual mode")
            slimbook.qc71.manual_control_set(True)
            
            #qc71_module_thread = threading.Thread(
                #name='slimbook.service.qc71.module', target=qc71_module_worker)
            #qc71_module_thread.start()
        
        else:
            logger.warning("QC71 kernel module is not available!")
            
    elif (platform in keyboard_platforms):
        keyboard_thread = threading.Thread(
            name='slimbook.service.generic.keyboard', target=keyboard_worker)
        keyboard_thread.start()
    
    else:
        logger.warning("No event handler for this model!")
        
    cached_events = {}
    expect_upower_event = False
    ac = False
    restore_profile = slimbook.info.SLB_QC71_PROFILE_BALANCED
    
    while True:
       
        event = slb_events.get()
        now = time.time()
        
        logger.debug("event {0:04X}".format(event))
        
        cached = cached_events.get(event)
        
        if cached:
            delta =  now - cached
            if delta > 1.250:
                cached_events[event] = now
            else:
                logger.debug("ignored duplicated event {0:04X} ({1})".format(event,delta))
                continue
        else:
            cached_events[event] = now
        
        # no need to bother user with this event as it is already notified elsewhere
        #if (event == common.SLB_EVENT_AC_OFFLINE or event == common.SLB_EVENT_AC_ONLINE):
        #   continue
        
        if (family == slimbook.info.SLB_MODEL_EXCALIBUR):
            if (event == common.SLB_EVENT_ENERGY_SAVER_MODE):
                set_power_profile(common.POWER_PROFILE_POWER_SAVER)
            elif (event == common.SLB_EVENT_BALANCED_MODE):
                set_power_profile(common.POWER_PROFILE_BALANCED)
            elif (event == common.SLB_EVENT_PERFORMANCE_MODE):
                set_power_profile(common.POWER_PROFILE_PERFORMANCE)

        if (platform == slimbook.info.SLB_PLATFORM_QC71):
            if (event == common.SLB_EVENT_QC71_INPUT_LOADED):
                qc71_module_thread = threading.Thread(
                    name='slimbook.service.qc71.module', target=qc71_module_worker)
                qc71_module_thread.start()
                
                module_loaded = True
                continue
                
            if (event == common.SLB_EVENT_QC71_INPUT_UNLOADED):
                module_loaded = False
                continue
                
            if (module_loaded):
                if (event == common.SLB_EVENT_QC71_SUPER_LOCK_CHANGED):
                    value = slimbook.qc71.super_lock_get()
                    if (value == 1):
                        event = common.SLB_EVENT_QC71_SUPER_LOCK_ON
                    else:
                        event = common.SLB_EVENT_QC71_SUPER_LOCK_OFF
                
                # General Performance event on QC71
                elif (event == common.SLB_EVENT_QC71_SILENT_MODE_CHANGED):
                    value = slimbook.qc71.profile_get()
                    logger.debug("current performance:{0}".format(common.POWER_PROFILE_NAME[value]))
                    
                    if (power_profiles == 2):
                        if (value == slimbook.info.SLB_QC71_PROFILE_SILENT):
                            slimbook.qc71.profile_set(slimbook.info.SLB_QC71_PROFILE_NORMAL)
                            logger.debug("switching to {0}".format(common.POWER_PROFILE_NAME[slimbook.info.SLB_QC71_PROFILE_NORMAL]))
                            event = common.SLB_EVENT_QC71_SILENT_MODE_OFF
                            expect_upower_event = True
                            set_power_profile(common.POWER_PROFILE_BALANCED)
                            
                        elif (value == slimbook.info.SLB_QC71_PROFILE_NORMAL):
                            slimbook.qc71.profile_set(slimbook.info.SLB_QC71_PROFILE_SILENT)
                            logger.debug("switching to {0}".format(common.POWER_PROFILE_NAME[slimbook.info.SLB_QC71_PROFILE_SILENT]))
                            event = common.SLB_EVENT_QC71_SILENT_MODE_ON
                            expect_upower_event = True
                            set_power_profile(common.POWER_PROFILE_POWER_SAVER)
                        
                    if (power_profiles == 3):
                        # Creative doesn't change TDP without AC'
                        if (ac == False and (family == slimbook.info.SLB_MODEL_CREATIVE)):
                            continue
                        
                        if (value == slimbook.info.SLB_QC71_PROFILE_PERFORMANCE):
                            slimbook.qc71.profile_set(slimbook.info.SLB_QC71_PROFILE_ENERGY_SAVER)
                            logger.debug("switching to {0}".format(common.POWER_PROFILE_NAME[slimbook.info.SLB_QC71_PROFILE_ENERGY_SAVER]))
                            event = common.SLB_EVENT_ENERGY_SAVER_MODE
                            expect_upower_event = True
                            set_power_profile(common.POWER_PROFILE_POWER_SAVER)
                            
                        elif (value == slimbook.info.SLB_QC71_PROFILE_ENERGY_SAVER):
                            slimbook.qc71.profile_set(slimbook.info.SLB_QC71_PROFILE_BALANCED)
                            logger.debug("switching to {0}".format(common.POWER_PROFILE_NAME[slimbook.info.SLB_QC71_PROFILE_BALANCED]))
                            event = common.SLB_EVENT_BALANCED_MODE
                            expect_upower_event = True
                            set_power_profile(common.POWER_PROFILE_BALANCED)
                            
                        elif (value == slimbook.info.SLB_QC71_PROFILE_BALANCED):
                            slimbook.qc71.profile_set(slimbook.info.SLB_QC71_PROFILE_PERFORMANCE)
                            logger.debug("switching to {0}".format(common.POWER_PROFILE_NAME[slimbook.info.SLB_QC71_PROFILE_PERFORMANCE]))
                            event = common.SLB_EVENT_PERFORMANCE_MODE
                            expect_upower_event = True
                            set_power_profile(common.POWER_PROFILE_PERFORMANCE)

                elif (event == common.SLB_EVENT_AC_OFFLINE):
                    ac = False
                    logger.debug("AC Offline")
                    if ((family == slimbook.info.SLB_MODEL_CREATIVE or family == slimbook.info.SLB_MODEL_EVO) and module_loaded):
                        restore_profile = slimbook.qc71.profile_get()
                        slimbook.qc71.profile_set(slimbook.info.SLB_QC71_PROFILE_BALANCED)
                        slimbook.qc71.profile_set(slimbook.info.SLB_QC71_PROFILE_ENERGY_SAVER)

                        if (settings[common.OPT_AC_NOTIFICATIONS] == False):
                            continue

                        event = common.QC71_TRIPLE_PROFILE_TO_NOTIFICATION[slimbook.info.SLB_QC71_PROFILE_ENERGY_SAVER]
                        

                elif (event == common.SLB_EVENT_AC_ONLINE):
                    ac = True
                    
                    if ((family == slimbook.info.SLB_MODEL_CREATIVE or family == slimbook.info.SLB_MODEL_EVO) and module_loaded):
                        
                        # restore profile
                        logger.debug("AC Online")
                        slimbook.qc71.profile_set(restore_profile)

                        if (settings[common.OPT_AC_NOTIFICATIONS] == False):
                            continue

                        event = common.QC71_TRIPLE_PROFILE_TO_NOTIFICATION[restore_profile]
                        
                elif (event & 0xfff0 == common.SLB_EVENT_UPOWER_POWER_EVENT):
                    # power profile matching is disabled
                    if (settings[common.OPT_POWER_PROFILE] == False):
                        continue
                        
                    if (expect_upower_event):
                        logger.debug("Expected UPower event, nothing to do")
                        expect_upower_event = False
                        continue

                    # Creative doesn't change TDP without AC'
                    if (ac == False and (family == slimbook.info.SLB_MODEL_CREATIVE)):
                        continue

                    if (power_profiles == 3):
                        expect = common.QC71_TRIPLE_PROFILE_FROM_UPOWER[event]
                        logger.debug("external power event {0:04X}, expected {1:04X}".format(event,expect))
                        slimbook.qc71.profile_set(expect)
                        event = common.QC71_TRIPLE_PROFILE_TO_NOTIFICATION[expect]
                        
                    elif (power_profiles == 2):
                        expect = common.QC71_DOUBLE_PROFILE_FROM_UPOWER[event]
                        logger.debug("external power event {0:04X}, expected {1:04X}".format(event,expect))
                        slimbook.qc71.profile_set(expect)
                        event = common.QC71_DOUBLE_PROFILE_TO_NOTIFICATION[expect]

        if (event == common.SLB_EVENT_TOUCHPAD_CHANGED):
            if (not settings[common.OPT_TRACKPAD_LOCK]):
                continue
            
            if (tpad.valid()):
                tpad.toggle()
                state = tpad.get_state()
                
                if (state == touchpad.Touchpad.STATE_LOCKED):
                    event = common.SLB_EVENT_TOUCHPAD_OFF
                elif (state == touchpad.Touchpad.STATE_UNLOCKED):
                    event = common.SLB_EVENT_TOUCHPAD_ON
                else:
                    continue
                
            else:
                #discard event
                continue
                    
        logger.debug("out event {0:04X}".format(event))
        send_notify(event)
        
if __name__=="__main__":
    try:
        main()
    except KeyboardInterrupt as e:
        sys.exit(0)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

# Slimbook Service
# Copyright (C) 2022 Slimbook
# In case you modify or redistribute this code you must keep the copyright line above.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import iohid

import evdev
import os

BUTTON_SWITCH_USAGE_ID = (iohid.HID_USAGE_PAGE_DIGITIZER << 16) | iohid.HID_USAGE_DIGITIZER_BUTTON_SWITCH
SURFACE_SWITCH_USAGE_ID = (iohid.HID_USAGE_PAGE_DIGITIZER << 16) | iohid.HID_USAGE_DIGITIZER_SURFACE_SWITCH

class Touchpad:
    MODE_UNKNOWN = 0
    MODE_HIDRAW = 1
    MODE_EVDEV = 2
    
    STATE_UNKNOWN = 0
    STATE_LOCKED = 1
    STATE_UNLOCKED = 2
    
    def __init__(self):
        self.mode = Touchpad.MODE_UNKNOWN
        self.report_id = 0
        self.fd = 0
        self.state = Touchpad.STATE_UNKNOWN
        self.device = None
        
        for device in iohid.list_devices():
            fd = os.open(device,os.O_RDWR)
            info = iohid.get_device_info(fd)
            found = False
        
            # ProX/Executive touchpad
            if (info.bus == iohid.HID_BUS_I2C and info.vendor == 0x93A):
                report = iohid.get_report_descriptor(fd)
                reports = iohid.parse_report_descriptor(report)
                
                for r in reports:
                    if r.report_type == iohid.HID_MAIN_FEATURE:
                        button_switch = False
                        surface_switch = False
                        
                        for usage in r.usages:
                            
                            if usage == BUTTON_SWITCH_USAGE_ID:
                                button_switch = True
                            if usage == SURFACE_SWITCH_USAGE_ID:
                                surface_switch = True
                        
                        if button_switch and surface_switch:
                            self.report_id = r.id
                            self.fd = fd
                            self.mode = Touchpad.MODE_HIDRAW
                            found = True
                if not found:
                    os.close(fd)
            else:
                os.close(fd)
        
            if found:
                try:
                    self.get_state()
                    return
                except:
                    #some devices fails with a errno 22 (Invalid Argument)
                    #better to fallback to evdev grab method
                    found = False
    
        found = False
        # generic touchpad lookup based on evdev grabing
        for devpath in evdev.list_devices():
            device = evdev.InputDevice(devpath)
    
            for cap in device.capabilities():
                # EV_KEY
                if cap==1:
                    for k in device.capabilities()[cap]:
                        #BTN_TOUCH
                        if k==330:
                            self.device = device
                            self.mode = Touchpad.MODE_EVDEV
                            
                            found = True
                            break
            
            if found:
                self.state = Touchpad.STATE_UNLOCKED
                break
                
    def lock(self):
        if self.mode == Touchpad.MODE_HIDRAW and self.fd>0:
            iohid.set_feature(self.fd,self.report_id,bytes([0x00]))
        
        if self.mode == Touchpad.MODE_EVDEV and self.device:
            self.device.grab()
            self.state = Touchpad.STATE_LOCKED
    
    def unlock(self):
        if self.mode == Touchpad.MODE_HIDRAW and self.fd>0:
            iohid.set_feature(self.fd,self.report_id,bytes([0x03]))
            
        if self.mode == Touchpad.MODE_EVDEV and self.device:
            self.device.ungrab()
            self.state = Touchpad.STATE_UNLOCKED
    
    def toggle(self):
        self.get_state()
    
        if (self.mode == Touchpad.MODE_HIDRAW and self.fd>0) or (self.mode == Touchpad.MODE_EVDEV and self.device):
            if self.state == Touchpad.STATE_LOCKED:
                
                self.unlock()
            elif self.state == Touchpad.STATE_UNLOCKED:
                
                self.lock()
            elif self.state == Touchpad.STATE_UNKNOWN:
                self.unlock()
        
    def get_state(self):
        
        if self.mode == Touchpad.MODE_HIDRAW and self.fd>0:
            self.state = Touchpad.MODE_UNKNOWN
            data = iohid.get_feature(self.fd, self.report_id,1)
            # mask is hardcoded, in the future maybe would be
            # better to obtain it from report descriptor
            data = int(data[0]) & 0x03
            
            if (data == 0):
                self.state = Touchpad.STATE_LOCKED
            else:
                self.state = Touchpad.STATE_UNLOCKED
        
        return self.state
        
    def valid(self):
        return (self.mode != Touchpad.MODE_UNKNOWN)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

# Slimbook Service
# Copyright (C) 2022 Slimbook
# In case you modify or redistribute this code you must keep the copyright line above.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from common import Configuration
from common import _

import slimbook.info

import zmq
import feedparser
import gi

import logging
import threading
import subprocess
import os
import sys
import shutil
import common
import webbrowser
import hashlib
import time
import signal
import fnmatch
import datetime
from dateutil import parser


try:
    gi.require_version("Gtk", "3.0")
    gi.require_version("Gio", "2.0")
    gi.require_version("GLib", "2.0")
    gi.require_version("GdkPixbuf", "2.0")
    gi.require_version("Notify", "0.7")

    try:
        gi.require_version("AyatanaAppIndicator3", "0.1")
        from gi.repository import AyatanaAppIndicator3 as appindicator
    except:
        gi.require_version("AppIndicator3", "0.1")
        from gi.repository import AppIndicator3 as appindicator
except Exception as e:
    print(e)
    exit(1)

from gi.repository import GObject
from gi.repository import Gtk, Gdk, Gio
from gi.repository import GLib
from gi.repository import GdkPixbuf
from gi.repository import Notify
from gi.repository import GLib

Notify.init("Slimbok Client Notifications")
notification = Notify.Notification.new("", "")
notification.set_app_name("Slimbok Client Notifications")
notification.set_timeout(Notify.EXPIRES_DEFAULT)
notification.set_urgency(Notify.Urgency.NORMAL)

logging.basicConfig(
    level=logging.INFO,
    format="[%(levelname)s] (%(threadName)-10s) %(message)s",
)

zmq_context = zmq.Context()


def update_server_settings(settings):
    logging.info("Updating server settings...")
    socket = zmq_context.socket(zmq.REQ)
    socket.connect("ipc://{0}".format(common.SLB_IPC_CTL_PATH))

    data = {}
    data["cmd"] = common.CMD_LOAD_SETTINGS
    data["settings"] = settings
    socket.send_json(data)
    socket.close()


class Feed:
    def __init__(self, entry):
        try:
            m = hashlib.md5()
            m.update(str(entry).encode())

            self.id = m.hexdigest()
            self.title = entry.title
            self.body = entry.description
            self.link = entry.get("link")
            self.published = entry.get("published")
            # print("link:",entry.get("link"))
            self.tags = []
            self.icon = "dialog-information"

            self.cached = False
            self.old = False

            if self.published:
                now = datetime.datetime.now(datetime.timezone.utc)
                ptime = parser.parse(self.published)
                delta = now - ptime
                self.old = delta.days > 90
            else:
                self.old = True

            if entry.get("tags"):
                for tag in entry.tags:
                    term = tag.get("term")

                    if term:
                        self.tags.append(term)

                        if term == "firmware":
                            self.icon = "application-x-firmware"

        except Exception as e:
            print(e)


def load_cache_feeds():
    feeds = []

    try:
        cache_file = os.path.expanduser("~/.cache/slimbook-service/feeds.dat")
        f = open(cache_file, "r")
        for line in f.readlines():
            value = line.strip()
            feeds.append(value)
        f.close()
    except:
        pass

    return feeds


def store_cache_feeds(feeds):
    try:
        cache_path = os.path.expanduser("~/.cache/slimbook-service/")
        os.makedirs(cache_path, exist_ok=True)

        f = open(cache_path + "feeds.dat", "w")
        for feed in feeds:
            f.write(feed.id + "\n")

        f.close()
    except Exception as e:
        print(e)


def check_time_feeds():
    feed = os.path.expanduser("~/.cache/slimbook-service/sb-rss.xml")

    if os.path.exists(feed):
        mtime = os.path.getmtime(feed)
        now = time.time()
        return (now - mtime) < (3600)
    else:
        return False


class ServiceIndicator(GObject.Object):
    def __init__(self):
        super().__init__()

        GObject.signal_new(
            "preferences-close",
            PreferencesDialog,
            GObject.SignalFlags.RUN_LAST,
            None,
            (GObject.TYPE_BOOLEAN,),
        )

        GObject.signal_new(
            "feed-update-start",
            ServiceIndicator,
            GObject.SignalFlags.RUN_LAST,
            None,
            (GObject.TYPE_BOOLEAN,),
        )

        GObject.signal_new(
            "feed-update-complete",
            ServiceIndicator,
            GObject.SignalFlags.RUN_LAST,
            None,
            (GObject.TYPE_BOOLEAN,),
        )

        # set up zmq
        self.socket = zmq_context.socket(zmq.SUB)
        self.socket.connect("ipc://{0}".format(common.SLB_IPC_PATH))
        self.socket.setsockopt_string(zmq.SUBSCRIBE, "")
        self.poller = zmq.Poller()
        self.poller.register(self.socket, zmq.POLLIN)

        GLib.idle_add(self.zmq_loop)

        self.set_indicator()
        Notify.init("Slimbook")

        GLib.timeout_add_seconds(5, self.on_notifications_timeout)

        self.feed_updating = False

    def zmq_loop(self):
        while self.poller.poll(timeout=50):
            data = self.socket.recv_json()
            code = data.get("code")
            event = common.SLB_EVENT_DATA.get(code)
            # avoid crashing on unhandled event codes
            if event is None:
                continue

            self.message("Slimbook", event[0], event[1])

        return True

    def on_notifications_timeout(self):
        if self.notifications_enabled:
            if not check_time_feeds():
                self.update_feed()

        GLib.timeout_add_seconds(3600 * 6, self.on_notifications_timeout)

        return False

    def update_feed(self):
        logging.info("updating feed...")

        if self.feed_updating == False:
            self.emit("feed-update-start", False)
            self.feed_updating = True
            thread = threading.Thread(target=self.update_feed_worker)
            thread.daemon = True
            thread.start()

    def update_feed_worker(self):
        try:
            common.download_feed()
            GLib.idle_add(self.on_feed_update)
        except:
            logging.warning("failed to get rss feed (no connection?)")

    def on_feed_update(self):
        logging.info("feed has been updated")
        self.feed_updating = False
        self.emit("feed-update-complete", False)

        if self.menu_news.get_sensitive():
            self.check_news()

    def check_news(self):
        news = []
        warn_user = False

        logging.info("checking news...")
        cached = load_cache_feeds()

        product = slimbook.info.product_name().lower().strip()
        sku = slimbook.info.product_sku().lower().strip()
        family = slimbook.info.get_family_name()
        ec_firmware = slimbook.info.ec_firmware_release()
        bios_version = slimbook.info.bios_version()
        logging.info("model:{0}".format(product))
        logging.info("sku:{0}".format(sku))
        logging.info("family:{0}".format(family))
        logging.info("ec:{0}".format(ec_firmware))
        logging.info("bios:{0}".format(bios_version))

        try:
            feed = feedparser.parse(
                os.path.expanduser("~/.cache/slimbook-service/sb-rss.xml")
            )
            now = datetime.datetime.now(datetime.timezone.utc)

            for entry in feed["entries"]:
                nw = Feed(entry)
                filters = 0
                match = False

                for tag in nw.tags:
                    if tag.startswith("family:"):
                        target = tag.split(":")[1]
                        filters = filters + 1
                        if fnmatch.fnmatch(family, target):
                            logging.info(
                                "feed match family filter:{0}={1}".format(
                                    family, target
                                )
                            )
                            match = True

                    if tag.startswith("model:"):
                        target = tag.split(":")[1]
                        filters = filters + 1

                        if fnmatch.fnmatch(product, target):
                            logging.info(
                                "feed match product filter:{0}={1}".format(
                                    product, target
                                )
                            )
                            match = True
                        elif fnmatch.fnmatch(sku, target):
                            logging.info(
                                "feed match sku filter:{0}={1}".format(sku, target)
                            )
                            match = True

                if filters > 0 and match == False:
                    logging.info("entry ignored by filter")
                    continue

                for cid in cached:
                    if cid == nw.id:
                        logging.info("id cached:{0}".format(nw.id))
                        nw.cached = True
                        break

                news.append(nw)

                body = nw.body

                if nw.link:
                    body = body + " " + nw.link

                if nw.cached == False and nw.old == False:
                    nt = Notify.Notification.new(nw.title, body, nw.icon)
                    nt.show()

                    warn_user = True

            store_cache_feeds(news)

        except Exception as e:
            logging.error(e)

        if warn_user:
            self.indicator.set_status(appindicator.IndicatorStatus.ATTENTION)
        else:
            self.indicator.set_status(
                appindicator.IndicatorStatus.ACTIVE
            ) if self.show else self.indicator.set_status(
                appindicator.IndicatorStatus.PASSIVE
            )

        return news

    def set_indicator(self):
        logging.debug("Setting indicator...")
        self.active_icon = None
        self.attention_icon = None
        self.about_dialog = None
        self.active = False

        self.notification = Notify.Notification.new("", "", None)
        self.read_preferences()
        manage_autostart(self.autostart)

        self.indicator = appindicator.Indicator.new(
            "com.slimbook.service",
            self.active_icon,
            appindicator.IndicatorCategory.HARDWARE,
        )

        self.indicator.set_title("Slimbook Client Notifications")

        self.indicator.set_attention_icon_full(self.attention_icon, "")

        self.running = True

        self.menu = self.get_menu()
        self.indicator.set_menu(self.menu)
        self.indicator.set_status(
            appindicator.IndicatorStatus.ACTIVE
        ) if self.show else self.indicator.set_status(
            appindicator.IndicatorStatus.PASSIVE
        )

    def message(self, title, message, icon="dialog-information"):
        notification.update(title, message, icon)
        notification.show()

    def read_preferences(self):
        configuration = Configuration()
        self.first_time = configuration.get("first-time")
        self.autostart = configuration.get("autostart")
        self.version = configuration.get("version")
        self.theme = configuration.get("theme")
        self.active_icon = common.STATUS_ICON[configuration.get("theme")]
        self.attention_icon = common.STATUS_ICON[self.theme + "-attention"]
        self.show = configuration.get("show")
        self.notifications_enabled = configuration.get("notifications")

        # push settings to server
        settings = {}

        for key in [common.OPT_TRACKPAD_LOCK, common.OPT_POWER_PROFILE]:
            value = configuration.get(key)
            if value is not None:
                settings[key] = value

        update_server_settings(settings)

    def get_menu(self):
        """Create and populate the menu."""
        menu = Gtk.Menu()

        self.menu_news = Gtk.MenuItem.new_with_label(_("Notifications"))
        self.menu_news.connect("activate", self.on_news_item)
        self.menu_news.show()
        menu.append(self.menu_news)

        menu_sysinfo = Gtk.MenuItem.new_with_label(_("System information"))
        menu_sysinfo.connect("activate", self.on_sysinfo_item)
        menu_sysinfo.show()
        menu.append(menu_sysinfo)

        self.menu_preferences = Gtk.MenuItem.new_with_label(_("Preferences"))
        self.menu_preferences.connect("activate", self.on_preferences_item)
        self.menu_preferences.show()
        menu.append(self.menu_preferences)

        if (os.path.exists(common.CONTROL_PANEL_PATH)):
            self.menu_control_panel = Gtk.MenuItem.new_with_label(_("Control Panel"))
            self.menu_control_panel.connect("activate", self.on_control_panel_item)
            self.menu_control_panel.show()
            menu.append(self.menu_control_panel)

        separator = Gtk.SeparatorMenuItem()
        separator.show()
        menu.append(separator)

        about_item = Gtk.MenuItem.new_with_label(_("About"))
        about_item.connect("activate", self.on_about_item)
        about_item.show()
        menu.append(about_item)

        self.report = Gtk.MenuItem.new_with_label(_("Generate report"))
        self.report.connect("activate", self.on_report_item)
        self.report.show()
        menu.append(self.report)

        bug_item = Gtk.MenuItem(label=_("Report a bug..."))
        bug_item.connect(
            "activate",
            lambda x: webbrowser.open(
                "https://github.com/slimbook/slimbook_service/issues/new"
            ),
        )
        bug_item.show()
        menu.append(bug_item)

        separator2 = Gtk.SeparatorMenuItem()
        separator2.show()
        menu.append(separator2)

        menu_exit = Gtk.MenuItem.new_with_label(_("Exit"))
        menu_exit.connect("activate", self.on_quit_item)
        menu_exit.show()
        menu.append(menu_exit)

        menu.show()

        return menu

    def get_about_dialog(self):
        """Create and populate the about dialog."""
        about_dialog = Gtk.AboutDialog()
        about_dialog.set_name(common.APPNAME)
        about_dialog.set_version(common.VERSION)
        about_dialog.set_copyright("Copyrignt (c) 2024\nSlimbook")
        about_dialog.set_comments(_("Slimbook Service"))
        about_dialog.set_license("""
This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
""")
        about_dialog.set_website("http://www.slimbook.es")
        about_dialog.set_website_label(_("Visit Website"))
        link = Gtk.LinkButton(
            uri=("https://github.com/slimbook/slimbook_service/issues/new"),
            label=(_("Report issue")),
        )
        link.set_name("link")
        link.set_halign(Gtk.Align.CENTER)
        about_dialog.set_authors(["Slimbook <dev@slimbook.es>","Antonio Masiá <ajmasia.dev@ysnp.link>"])
        about_dialog.set_documenters(["Slimbook <dev@slimbook.es>"])
        about_dialog.set_translator_credits(_("translator-credits"))

        theme = Gtk.IconTheme()
        pix = theme.load_icon(
            icon_name=common.ICON, size=125, flags=Gtk.IconLookupFlags.FORCE_SYMBOLIC
        )
        about_dialog.set_icon(pix)
        about_dialog.set_logo(pix)
        about_dialog.set_program_name(common.APPNAME)
        return about_dialog

    def on_preferences_item(self, widget, data=None):
        self.show_preferences()

    def on_control_panel_item(self, widget, data=None):
        subprocess.Popen([common.CONTROL_PANEL_PATH])

    def on_sysinfo_item(self, widget, data=None):
        logging.debug("system info")
        widget.set_sensitive(False)
        info = common.get_system_info()

        sysinfo_dialog = SystemInfoDialog(info)
        sysinfo_dialog.run()
        sysinfo_dialog.destroy()
        widget.set_sensitive(True)

    def on_news_item(self, widget, data=None):
        logging.debug("news")
        widget.set_sensitive(False)
        news_dialog = NotificationsDialog(self)
        news_dialog.connect("delete-event", self.on_news_delete_event)

    def on_quit_item(self, widget, data=None):
        Notify.uninit()
        logging.debug("Exit")
        exit(0)

    def on_about_item(self, widget, data=None):
        if self.about_dialog:
            self.about_dialog.present()
        else:
            self.about_dialog = self.get_about_dialog()
            self.about_dialog.run()
            self.about_dialog.destroy()
            self.about_dialog = None

    def on_report_item(self, widget, data=None):
        self.show_report()
        widget.set_sensitive(True)

    # Interface and Method

    def on_news_delete_event(self, window, event):
        self.menu_news.set_sensitive(True)
        self.indicator.set_status(
            appindicator.IndicatorStatus.ACTIVE
        ) if self.show else self.indicator.set_status(
            appindicator.IndicatorStatus.PASSIVE
        )

    def on_preferences_close(self, *args):
        self.menu_preferences.set_sensitive(True)
        self.read_preferences()

        self.indicator.set_attention_icon_full(self.attention_icon, "")
        self.indicator.set_icon_full(self.active_icon, "")

    def show_preferences(self):
        self.menu_preferences.set_sensitive(False)
        preferences_dialog = PreferencesDialog()
        preferences_dialog.connect("preferences-close", self.on_preferences_close)

    def show_report(self):
        self.report.set_sensitive(False)
        report_dialog = ReportDialog()


class ReportDialog(Gtk.Window):
    def __init__(self):
        Gtk.Window.__init__(self)
        self.set_default_size(600, 100)
        self.set_modal(True)

        self.path = ""
        self.has_ended = False

        self.connect("delete-event", self.on_report_delete_event)

        self.set_position(Gtk.WindowPosition.CENTER_ALWAYS)
        theme = Gtk.IconTheme()
        pix = theme.load_icon(
            icon_name=common.ICON, size=64, flags=Gtk.IconLookupFlags.FORCE_SYMBOLIC
        )
        self.set_icon(pix)

        header = Gtk.HeaderBar()
        header.set_title(_("Generate report"))
        header.set_show_close_button(True)

        self.set_titlebar(header)

        self.stack = Gtk.Stack()

        vboxrv = Gtk.VBox()
        vboxmv = Gtk.VBox()
        vboxev = Gtk.VBox()

        # Report View

        vboxrv.set_margin_start(20)
        vboxrv.set_margin_end(20)
        vboxrv.set_margin_top(10)
        vboxrv.set_margin_bottom(10)

        hboxrv = Gtk.HBox()
        hboxrv.set_margin_start(5)
        hboxrv.set_margin_end(5)

        report_desc = Gtk.Label.new(
            _(
                "This is a report of several hardware and software stats.\nFull report generates a report with sensitive information,\nbeware of sharing it online!"
            )
        )

        vboxrv.pack_start(report_desc, True, True, 4)

        self.progress_bar = Gtk.ProgressBar()
        self.progress_bar.set_text("")
        self.progress_bar.set_show_text(True)

        vboxrv.pack_start(self.progress_bar, True, True, 4)

        self.normal_report_btn = Gtk.Button.new_with_label(_("Report"))
        self.normal_report_btn.connect("clicked", self.on_report_button)

        self.full_report_btn = Gtk.Button.new_with_label(_("Full report"))
        self.full_report_btn.connect("clicked", self.on_full_report_button)

        hboxrv.pack_start(self.normal_report_btn, True, True, 4)
        hboxrv.pack_start(self.full_report_btn, True, True, 4)

        vboxrv.pack_start(hboxrv, True, True, 4)

        # Message View

        vboxmv.set_margin_start(20)
        vboxmv.set_margin_end(20)
        vboxmv.set_margin_top(10)
        vboxmv.set_margin_bottom(10)

        hboxmv = Gtk.HBox()
        hboxmv.set_margin_start(10)
        hboxmv.set_margin_end(10)
        hboxmv.set_margin_top(15)
        hboxmv.set_margin_bottom(15)

        self.path_label = Gtk.Label.new(self.path)
        vboxmv.pack_start(self.path_label, True, True, 4)

        self.open_btn = Gtk.Button.new_with_label(_("Open"))
        self.open_btn.connect("clicked", self.on_open_button)

        self.close_btn = Gtk.Button.new_with_label(_("Close"))
        self.close_btn.connect("clicked", self.on_close_button)

        hboxmv.pack_start(self.open_btn, True, True, 4)
        hboxmv.pack_start(self.close_btn, True, True, 4)

        vboxmv.pack_start(hboxmv, True, True, 4)

        # Error view

        vboxev.set_margin_start(20)
        vboxev.set_margin_end(20)
        vboxev.set_margin_top(10)
        vboxev.set_margin_bottom(10)

        hboxev = Gtk.HBox()
        hboxev.set_margin_start(10)
        hboxev.set_margin_end(10)
        hboxev.set_margin_top(15)
        hboxev.set_margin_bottom(15)

        self.err_code = ""

        self.err_code_label = Gtk.Label.new(self.err_code)

        vboxev.pack_start(self.err_code_label, True, True, 4)

        self.close_btn_err = Gtk.Button.new_with_label(_("Close"))
        self.close_btn_err.connect("clicked", self.on_close_button)

        hboxev.pack_start(self.close_btn_err, True, True, 4)

        vboxev.pack_start(hboxev, True, True, 4)

        self.stack.add_named(vboxrv, name="Report view")

        self.stack.add_named(vboxmv, name="Message view")

        self.stack.add_named(vboxev, name="Error view")

        self.add(self.stack)

        self.show_all()

    def prog_bar_proc(self, args):
        if args[0] == True:
            self.path_label.set_label(_("Succesful! Dumped at ") + self.path)
            self.resize(200, 100)
            self.stack.set_visible_child_name("Message view")
            self.progress_bar.set_fraction(1.0)
        else:
            self.progress_bar.pulse()

        if args[1] != "":
            self.err_code_label.set_label(
                _("Error! Report wasn't able to be generated\nError: ") + self.err_code
            )
            self.stack.set_visible_child_name("Error view")
            self.resize(200, 100)
            self.err_code = args[1]

        if args[2] != "":
            self.path = args[2]

    def on_report_button_common(self, widget, str):
        self.bar_thread = ReportThread(self.prog_bar_proc, str)
        self.bar_thread.start()
        self.disable_buttons()

    def on_report_button(self, widget):
        self.on_report_button_common(widget, "report")

    def on_full_report_button(self, widget):
        self.on_report_button_common(widget, "report-full")

    def disable_buttons(self):
        self.normal_report_btn.set_sensitive(False)
        self.full_report_btn.set_sensitive(False)

    def on_close_button(self, widget):
        self.close()

    def on_open_button(self, widget):
        subprocess.Popen(["xdg-open", os.path.dirname(self.path)])

    def on_report_delete_event(self, window, event):
        self.set_sensitive(False)


class ReportThread(threading.Thread):
    def __init__(self, cb, report_type):
        threading.Thread.__init__(self)
        self.callback = cb
        self.report_type = report_type

    def run(self):
        common.report_proc(self, GLib.idle_add, self.callback, self.report_type)


class PreferencesDialog(Gtk.Window):
    def __init__(self):
        Gtk.Window.__init__(self)
        self.set_modal(True)

        self.connect("delete-event", self.on_delete_event)

        self.set_position(Gtk.WindowPosition.CENTER_ALWAYS)
        theme = Gtk.IconTheme()
        pix = theme.load_icon(
            icon_name=common.ICON, size=64, flags=Gtk.IconLookupFlags.FORCE_SYMBOLIC
        )
        self.set_icon(pix)

        header = Gtk.HeaderBar()
        header.set_title(_("Slimbook Preferences"))
        header.set_show_close_button(True)

        self.btn_save = Gtk.Button.new_with_label(_("Save"))
        self.btn_save.set_sensitive(False)
        self.btn_save.connect("clicked", self.on_btn_save_clicked)
        header.pack_end(self.btn_save)
        self.set_titlebar(header)

        vbox0 = Gtk.VBox(spacing=5)
        vbox0.set_border_width(20)
        self.add(vbox0)
        table1 = Gtk.Table(n_rows=10, n_columns=2, homogeneous=False)
        vbox0.pack_start(table1, False, True, 1)

        label0 = Gtk.Label(label=_("Show indicator") + ":")
        label0.set_halign(Gtk.Align.CENTER)
        table1.attach(label0, 0, 1, 6, 7, xpadding=15, ypadding=15)
        self.switch0 = Gtk.Switch()
        table1.attach(
            self.switch0,
            1,
            2,
            6,
            7,
            xpadding=15,
            ypadding=15,
            xoptions=Gtk.AttachOptions.SHRINK,
        )

        label1 = Gtk.Label(label=_("Autostart") + ":")
        label1.set_halign(Gtk.Align.CENTER)
        table1.attach(label1, 0, 1, 7, 8, xpadding=15, ypadding=15)
        self.switch1 = Gtk.Switch()
        table1.attach(
            self.switch1,
            1,
            2,
            7,
            8,
            xpadding=15,
            ypadding=15,
            xoptions=Gtk.AttachOptions.SHRINK,
        )

        label2 = Gtk.Label(label=_("Light-mode Icon") + ":")
        label2.set_halign(Gtk.Align.CENTER)
        table1.attach(label2, 0, 1, 8, 9, xpadding=15, ypadding=15)
        self.switch2 = Gtk.Switch()
        table1.attach(
            self.switch2,
            1,
            2,
            8,
            9,
            xpadding=15,
            ypadding=15,
            xoptions=Gtk.AttachOptions.SHRINK,
        )

        label3 = Gtk.Label(label=_("Check Notifications") + ":")
        label3.set_halign(Gtk.Align.CENTER)
        table1.attach(label3, 0, 1, 9, 10, xpadding=15, ypadding=15)
        self.switch3 = Gtk.Switch()
        table1.attach(
            self.switch3,
            1,
            2,
            9,
            10,
            xpadding=15,
            ypadding=15,
            xoptions=Gtk.AttachOptions.SHRINK,
        )

        label4 = Gtk.Label(label=_("Trackpad lock") + ":")
        label4.set_halign(Gtk.Align.CENTER)
        table1.attach(label4, 0, 1, 10, 11, xpadding=15, ypadding=15)
        self.switch4 = Gtk.Switch()
        table1.attach(
            self.switch4,
            1,
            2,
            10,
            11,
            xpadding=15,
            ypadding=15,
            xoptions=Gtk.AttachOptions.SHRINK,
        )

        label5 = Gtk.Label(label=_("Set power profile") + ":")
        label5.set_halign(Gtk.Align.CENTER)
        table1.attach(label5, 0, 1, 11, 12, xpadding=15, ypadding=15)
        self.switch5 = Gtk.Switch()
        table1.attach(
            self.switch5,
            1,
            2,
            11,
            12,
            xpadding=15,
            ypadding=15,
            xoptions=Gtk.AttachOptions.SHRINK,
        )

        label6 = Gtk.Label(label=_("AC Notifications") + ":")
        label6.set_halign(Gtk.Align.CENTER)
        table1.attach(label6, 0, 1, 12, 13, xpadding=15, ypadding=15)
        self.switch6 = Gtk.Switch()
        table1.attach(
            self.switch6,
            1,
            2,
            12,
            13,
            xpadding=15,
            ypadding=15,
            xoptions=Gtk.AttachOptions.SHRINK,
        )

        self.load_preferences()

        self.changes = False
        self.switch0.connect("state-set", self.on_switch_state_set)
        self.switch1.connect("state-set", self.on_switch_state_set)
        self.switch2.connect("state-set", self.on_switch_state_set)
        self.switch3.connect("state-set", self.on_switch_state_set)
        self.switch4.connect("state-set", self.on_switch_state_set)
        self.switch5.connect("state-set", self.on_switch_state_set)
        self.switch6.connect("state-set", self.on_switch_state_set)

        self.show_all()

    def on_switch_state_set(self, switch, state):
        self.btn_save.set_sensitive(True)
        self.changes = True

    def on_delete_event(self, window, event):
        self.emit("preferences-close", self.changes)
        return False

    def on_btn_save_clicked(self, widget):
        self.save_preferences()
        self.btn_save.set_sensitive(False)

    def close_ok(self):
        self.save_preferences()

    def load_preferences(self):
        configuration = Configuration()
        first_time = configuration.get("first-time")
        version = configuration.get("version")
        if first_time or version != common.VERSION:
            configuration.set_defaults()
            configuration.read()

        self.switch0.set_active(configuration.get("show") == True)
        self.switch1.set_active(os.path.exists(common.FILE_AUTO_START))
        self.switch2.set_active(configuration.get("theme") == "light")
        self.switch3.set_active(configuration.get("notifications") == True)
        self.switch4.set_active(configuration.get("trackpad-lock") == True)
        self.switch5.set_active(configuration.get("power-profile") == True)
        self.switch6.set_active(configuration.get("ac-notifications") == True)

    def save_preferences(self):
        configuration = Configuration()
        configuration.set("first-time", False)
        configuration.set("version", common.VERSION)
        configuration.set("show", self.switch0.get_active())

        manage_autostart(self.switch1.get_active())
        if self.switch2.get_active():
            configuration.set("theme", "light")
        else:
            configuration.set("theme", "dark")

        configuration.set("notifications", self.switch3.get_active())
        configuration.set("trackpad-lock", self.switch4.get_active())
        configuration.set("power-profile", self.switch5.get_active())
        configuration.set("ac-notifications", self.switch6.get_active())
        configuration.save()

        settings = {}
        settings[common.OPT_TRACKPAD_LOCK] = self.switch4.get_active()
        settings[common.OPT_POWER_PROFILE] = self.switch5.get_active()
        settings[common.OPT_AC_NOTIFICATIONS] = self.switch6.get_active()
        update_server_settings(settings)


class SystemInfoDialog(Gtk.Dialog):
    def __init__(self, info):
        Gtk.Dialog.__init__(
            self,
            _("Slimbook System information"),
            None,
            modal=True,
            destroy_with_parent=True,
            use_header_bar=True,
        )

        CSS = """
            list {
                border-width: 1px;
                border-style: inset;
                border-color: lightgrey;
            }

            row {
                border-width: 1px;
                border-style: outset;
                border-color: lightgrey;
                min-height: 32px;
                min-width: 600px;
            }
            """
        self.resize(800, 700)
        self.info = info

        provider = Gtk.CssProvider()
        provider.load_from_data(CSS.encode("utf-8"))
        style_context = self.get_style_context()
        style_context.add_provider_for_screen(
            self.get_screen(), provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

        # btn_copy = Gtk.Button(label=_("Copy"))
        btn_copy = Gtk.Button.new_from_icon_name("edit-copy", Gtk.IconSize.BUTTON)
        btn_copy.connect("clicked", self.btn_copy_clicked)
        self.get_header_bar().pack_end(btn_copy)

        scrw = Gtk.ScrolledWindow()
        scrw.set_border_width(12)
        listbox = Gtk.ListBox()
        listbox.set_vexpand(True)
        listbox.set_hexpand(True)

        listbox.set_selection_mode(Gtk.SelectionMode.NONE)

        scrw.add(listbox)
        self.get_content_area().add(scrw)

        for k in info:
            key = k[0]
            value = k[1]
            label_key = Gtk.Label(label=key)
            # label_key.set_markup("<b>{0}</b>".format(key))
            label_value = Gtk.Label(label=value)

            hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)

            hbox.pack_start(label_key, False, False, 1)
            hbox.pack_end(label_value, False, False, 1)

            row = Gtk.ListBoxRow()
            row.add(hbox)

            listbox.add(row)

        self.show_all()

    def btn_copy_clicked(self, button):
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)

        txt = ""
        for k in self.info:
            key = k[0]
            value = k[1]
            txt = txt + "{0}:\t{1}\n".format(key, value)

        clipboard.set_text(txt, -1)
        button.set_sensitive(False)


class NotificationsDialog(Gtk.Window):
    def __init__(self, parent):
        Gtk.Window.__init__(self)
        self.set_modal(True)
        self.parent = parent
        self.set_default_size(500, 600)

        CSS = """
            list {
                border-width: 1px;
                border-style: inset;
                border-color: lightgrey;
            }

            row {
                border-width: 1px;
                border-style: outset;
                border-color: lightgrey;
                min-height: 32px;
                min-width: 400px;
            }
            """

        provider = Gtk.CssProvider()
        provider.load_from_data(CSS.encode("utf-8"))
        style_context = self.get_style_context()
        style_context.add_provider_for_screen(
            self.get_screen(), provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

        parent.connect("feed-update-start", self.on_feed_update_start)
        parent.connect("feed-update-complete", self.on_feed_update_complete)

        header = Gtk.HeaderBar()
        header.set_title(_("Slimbook Notifications"))
        header.set_show_close_button(True)

        self.btn_refresh = Gtk.Button.new_with_label(_("Refresh"))
        self.btn_refresh.connect("clicked", self.on_btn_refresh_clicked)
        header.pack_end(self.btn_refresh)

        self.set_titlebar(header)

        vbox = Gtk.VBox(spacing=12)
        sw = Gtk.ScrolledWindow()
        self.listbox = Gtk.ListBox()

        self.listbox.set_selection_mode(Gtk.SelectionMode.NONE)

        self.add(vbox)
        sw.add(self.listbox)
        vbox.pack_start(sw, True, True, 1)
        vbox.set_border_width(16)

        self.populate()

        self.show_all()

    def populate(self):
        # feeds = check_news()
        feeds = self.parent.check_news()

        for feed in feeds:
            grid = Gtk.Grid.new()
            grid.set_row_spacing(4)
            grid.set_column_spacing(8)

            lbl_title = Gtk.Label()
            lbl_title.set_markup("<b>{0}</b>".format(feed.title))
            lbl_title.set_halign(Gtk.Align.START)

            lbl_body = Gtk.Label(label=feed.body)
            lbl_body.set_halign(Gtk.Align.START)
            if feed.link:
                btn_link = Gtk.LinkButton(uri=feed.link, label=feed.link)
                btn_link.set_halign(Gtk.Align.START)
                grid.attach(btn_link, 1, 2, 1, 1)

            theme = Gtk.IconTheme()
            pix = theme.load_icon(
                icon_name=feed.icon, size=32, flags=Gtk.IconLookupFlags.FORCE_SYMBOLIC
            )

            img = Gtk.Image.new_from_pixbuf(pix)

            grid.attach(img, 0, 0, 1, 4)

            grid.attach(lbl_title, 1, 0, 1, 1)
            grid.attach(lbl_body, 1, 1, 1, 1)

            row = Gtk.ListBoxRow()
            row.add(grid)

            self.listbox.add(row)

        if len(feeds) == 0:
            theme = Gtk.IconTheme()

            pix = theme.load_icon(
                icon_name="face-plain-symbolic",
                size=32,
                flags=Gtk.IconLookupFlags.FORCE_SYMBOLIC,
            )

            img = Gtk.Image.new_from_pixbuf(pix)
            lbl = Gtk.Label(label=_("Nothing to show"))

            grid = Gtk.Grid.new()
            grid.set_row_spacing(4)
            grid.set_column_spacing(8)
            grid.attach(img, 0, 0, 1, 4)
            grid.attach(lbl, 1, 1, 1, 1)

            row = Gtk.ListBoxRow()
            row.add(grid)

            self.listbox.add(row)

        self.listbox.show_all()

    def on_btn_refresh_clicked(self, widget):
        self.parent.update_feed()
        self.show_feed_update()

    def on_feed_update_start(self, *args):
        self.show_feed_update()

    def show_feed_update(self):
        self.btn_refresh.set_sensitive(False)
        children = self.listbox.get_children()
        for child in children:
            self.listbox.remove(child)

        theme = Gtk.IconTheme()

        pix = theme.load_icon(
            icon_name="emblem-synchronizing-symbolic",
            size=32,
            flags=Gtk.IconLookupFlags.FORCE_SYMBOLIC,
        )

        img = Gtk.Image.new_from_pixbuf(pix)
        lbl = Gtk.Label(label=_("Fetching..."))

        grid = Gtk.Grid.new()
        grid.set_row_spacing(4)
        grid.set_column_spacing(8)
        grid.attach(img, 0, 0, 1, 4)
        grid.attach(lbl, 1, 1, 1, 1)

        row = Gtk.ListBoxRow()
        row.add(grid)

        self.listbox.add(row)
        self.listbox.show_all()

    def on_feed_update_complete(self, *args):
        self.btn_refresh.set_sensitive(True)
        children = self.listbox.get_children()

        for child in children:
            self.listbox.remove(child)
        self.populate()


def manage_autostart(create):
    if not os.path.exists(common.AUTOSTART_DIR):
        os.makedirs(common.AUTOSTART_DIR)
    if create:
        if not os.path.exists(common.FILE_AUTO_START):
            shutil.copyfile(common.FILE_AUTO_START_ORIG, common.FILE_AUTO_START)
    else:
        if os.path.exists(common.FILE_AUTO_START):
            os.remove(common.FILE_AUTO_START)


PID_FILE = os.path.join(
    os.environ.get("XDG_RUNTIME_DIR", "/tmp"), "slimbook-service-indicator.pid"
)

_indicator_instance = None


def _on_sigusr1(*args):
    if _indicator_instance:
        GLib.idle_add(_indicator_instance.show_preferences)


def is_running():
    try:
        with open(PID_FILE, "r") as f:
            pid = int(f.read().strip())
        os.kill(pid, 0)
        return True
    except (FileNotFoundError, ValueError, ProcessLookupError, PermissionError):
        return False


def send_preferences_signal():
    try:
        with open(PID_FILE, "r") as f:
            pid = int(f.read().strip())
        os.kill(pid, signal.SIGUSR1)
        return True
    except (FileNotFoundError, ValueError, ProcessLookupError, PermissionError):
        return False


def init_indicator():
    global _indicator_instance
    try:
        _indicator_instance = ServiceIndicator()

        with open(PID_FILE, "w") as f:
            f.write(str(os.getpid()))

        signal.signal(signal.SIGUSR1, _on_sigusr1)
        GLib.MainLoop().run()
    except KeyboardInterrupt:
        logging.info("out of main loop")
    finally:
        try:
            os.remove(PID_FILE)
        except OSError:
            pass
        exit(0)


def main():
    if "--preferences" in sys.argv or "-p" in sys.argv:
        if send_preferences_signal():
            return
    elif is_running():
        return
    init_indicator()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Slimbook Service
# Copyright (C) 2022 Slimbook
# In case you modify or redistribute this code you must keep the copyright line above.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import glob
import struct
from fcntl import ioctl

HID_MAX_DESCRIPTOR_SIZE = 4096

HIDIOCSFEATURE = 0xC0004806
HIDIOCGFEATURE = 0xC0004807
HIDIOCGRDESCSIZE = 0x80044801
HIDIOCGRDESC = 0x90044802
HIDIOCGRAWINFO = 0x80084803

HID_BUS_USB = 0x03
HID_BUS_HIL = 0x04
HID_BUS_BLUETOOTH = 0x05
HID_BUS_VIRTUAL = 0x06
HID_BUS_I2C = 0x18

HID_BUS = {
    HID_BUS_USB : "usb",
    HID_BUS_HIL : "hil",
    HID_BUS_BLUETOOTH : "bluetooth",
    HID_BUS_VIRTUAL : "virtual",
    HID_BUS_I2C : "i2c"
}

HID_SIZE = {
    0:0,
    1:1,
    2:2,
    3:4
    }

HID_TYPE_MAIN = 0
HID_TYPE_GLOBAL = 1
HID_TYPE_LOCAL = 2

HID_MAIN_INPUT = 0x80
HID_MAIN_OUTPUT = 0x90
HID_MAIN_FEATURE = 0xB0 
HID_MAIN_COLLECTION = 0xA0
HID_MAIN_END_COLLECTION = 0xC0

HID_LOCAL_USAGE = 0x08

HID_GLOBAL_USAGE_PAGE = 0x04
HID_GLOBAL_REPORT_ID = 0x84

HID_COLLECTION_PHYSICAL = 0x00
HID_COLLECTION_APPLICATION = 0x01
HID_COLLECTION_LOGICAL = 0x02

HID_USAGE_PAGE_DIGITIZER = 0x0D

HID_USAGE_DIGITIZER_DIGITIZER = 0x01
HID_USAGE_DIGITIZER_PEN = 0x02
HID_USAGE_DIGITIZER_DEVICE_CONFIGURATION = 0x0E
HID_USAGE_DIGITIZER_PEN = 0x22
HID_USAGE_DIGITIZER_DEVICE_MODE = 0x52
HID_USAGE_DIGITIZER_SURFACE_SWITCH = 0x57
HID_USAGE_DIGITIZER_BUTTON_SWITCH = 0x58

HID_COLLECTION = {
    0x00:"Physical",
    0x01:"Application",
    0x02:"Logical",
    0x03:"Report",
    0x04:"Named array",
    0x05:"Usage Switch",
    0x06:"Usage Mod"
    }
    
HID_MAIN = {
    0x80:"Input",
    0x90:"Output",
    0xB0:"Feature",
    0xA0:"Collection",
    0xC0:"End Collection"
    }

class DeviceInfo:
    def __init__(self,bus,vendor,product):
        self.bus = bus
        self.vendor = vendor
        self.product = product
        
    def __str__(self):
        return "{0} {1:04x} {2:04x}".format(HID_BUS[self.bus],self.vendor,self.product)

class Collection:
    def __init__(self,collection_type, usage_page, usage):
        self.collection_type = collection_type
        self.usage_page = usage_page
        self.usage = usage
        self.children = []
        
    def __str__(self):
        text = "Collection {0}, Usage Page {1:02x}, Usage {2:02x}\n".format(HID_COLLECTION[self.collection_type],self.usage_page,self.usage)
        for c in self.children:
            text = text + "...."+ str(c)
        
        return text
        
class Report:
    def __init__(self,id,report_type,usages):
        self.id = id
        self.report_type = report_type
        self.usages = usages
        
    def __str__(self):
        text = "{0} {1}\n".format(self.id,HID_MAIN[self.report_type])
        
        for u in self.usages:
            text = text + "...." + "{0:04x}\n".format(u)
        
        return text

def list_devices():
    return glob.glob("/dev/hidraw*")

def get_device_info(fd):
    info = struct.pack("Ihh",0,0,0)
    status = ioctl(fd,HIDIOCGRAWINFO,info)
    data = struct.unpack("Ihh",status)
    
    return DeviceInfo(data[0], data[1], data[2])

def set_feature(fd,id,data):
    cmd = HIDIOCSFEATURE | ((1 + len(data))<<16)
    data = bytes([id]) + data
    return ioctl(fd,cmd,data)

def get_feature(fd,id,size):
    cmd = HIDIOCGFEATURE | ((1 + size)<<16)
    data = bytes([id]) + bytes([0]*size)
    return ioctl(fd,cmd,data)[1:]

def get_report_descriptor(fd):
    data = struct.pack("I",0)
    status = ioctl(fd, HIDIOCGRDESCSIZE, data)
    report_size = struct.unpack("I",status)[0]
    
    report = struct.pack("I{0}s".format(report_size),report_size,bytes([0]*report_size))
    status = ioctl(fd, HIDIOCGRDESC, report)
    
    return struct.unpack("I{0}s".format(report_size),status)[1]

def parse_report_descriptor(data):
    n = 0
    report_size = len(data)
    usage_page = 0
    usage = 0
    report_id = 0
    reports = []
    usages = []
    
    while (n<report_size):
        byte = data[n]
        bSize = HID_SIZE[byte & 0x03]
        bType = (byte & 0x0C) >> 2
        bTag = (byte & 0xF0)
        
        if (bType == HID_TYPE_MAIN):
            main_type = byte & 0xFC
            
            if (main_type == HID_MAIN_COLLECTION):
                collection_type = data[n+1]

            if (main_type == HID_MAIN_END_COLLECTION):
                pass
                
            if (main_type == HID_MAIN_INPUT or main_type == HID_MAIN_OUTPUT or main_type == HID_MAIN_FEATURE):
                
                reports.append(Report(report_id,main_type,usages))
                usages = []
        
        if (bType == HID_TYPE_GLOBAL):
            global_type = byte & 0xFC

            if (global_type == HID_GLOBAL_USAGE_PAGE):
                usage_page = data[n+1]

            if (global_type == HID_GLOBAL_REPORT_ID):
                report_id = data[n+1]
        
        if (bType == HID_TYPE_LOCAL):
            local_type = byte & 0xFC

            if (local_type == HID_LOCAL_USAGE):
                usage = data[n+1]
                
                usages.append((usage_page<<16) | usage)
        
        n = n + bSize
        n = n + 1
    
    return reports

